(in-package :araneida)

;;; Make emacs evaluate
;;;  (put 'with-ignored-signals 'lisp-indent-function 'defun)
;;; for sane indenting in HANDLE-REQUEST

(defun accept-handler (listener)
  (let* ((socket (sockets:socket-accept listener)))
    (handle-request
     (sockets:socket-make-stream socket :element-type 'base-char
				 :input t :output t :buffering :full)
     socket)))

(defun install-serve-event-handlers ()
  (let* ((ports (remove-duplicates
                 (mapcar #'server-port *exported-servers*)))
         (listeners
          (mapcar (lambda (x)
                    (listen-to-inet-port :port x :reuse 1))
                  ports)))
    (unless listeners (error "No sockets to listen on"))
    (dolist (socket listeners)
      (event:add-fd-handler (sockets:socket-file-descriptor  socket)
			    :input (lambda (fd) (accept-handler socket))))))
				     
(defun read-new-sequence (length stream)
  ;; don't use READ-SEQUENCE, it's buggy
  (if (> length 0)
      (let ((buf
             (make-array length :element-type (stream-element-type stream))))
        (subseq buf 0 (event:read-n-bytes stream buf 0 length)))
    "" ))

(eval-when (:load-toplevel :execute)
  (sb-sys:ignore-interrupt :sigpipe))

(defun handle-request (stream socket)
  (let* ((request-line (read-folded-line stream nil))
         (headers (read-headers stream))
         (content-length (parse-integer
                          (or (header-value :content-length headers) "0")))
         (body (read-new-sequence content-length stream)))
    (destructuring-bind
        (method url-string protocol) (split request-line 3 '(#\Space))
      (declare (ignore protocol))
      ;; When we make the request, we ignore working out what the
      ;; prefix is.  dispatch-request will fill that in for us
      (let* ((parsed-body (if content-length (parse-body body) nil))
             (url (merge-url
                   (server-base-url (server-for-name-colon-port
                                     (header-value :host headers)))
                   url-string))
             (r (make-instance
                 'request :url url ; :original-url url
                 :method (intern (string-upcase method) 'keyword)
                 :body parsed-body
                 :stream stream :headers headers :socket socket)))
        ;; for each in :authentication :authorization :response :log
        ;; run the appropriate handler
        ;; if it caused an error, cut to :error
        ;; if it closed the stream, cut to :log
	(catch 'done
	  (handler-bind ((error
			  (lambda (c)
			    (format t "got error! r=~A c=~S~%" r c)
			    (setf (slot-value r 'condition) c)
			    (dispatch-request r :error)
			    (throw 'done nil))))
	    (progn
	      (dispatch-request r :authentication)
	      (unless (open-stream-p (request-stream r)) (throw 'done nil))
	      (dispatch-request r :authorization)
	      (unless (open-stream-p (request-stream r))  (throw 'done nil))
	      (dispatch-request r :response))))
	(let* ((s (request-stream r))
	       (fd (sb-sys:fd-stream-fd s)))
	  (multiple-value-bind (r e) (ignore-errors (close s) t)
	    (unless r
	      (format t "Unable to close fd ~A: ~A, trying harder ~%" fd e)
	      (multiple-value-bind (r e) (ignore-errors (close s :abort t) t)
		(unless r
		  (format t "still unable to close ~A: ~A, try harder ~%" fd e)
		  (multiple-value-bind (r e)
		      (ignore-errors (sb-unix:unix-close fd) t)
		    (unless r
		      (format t "Even unix-close failed on ~A:~A, giving up~%"
			      fd e)))))))
	  (dispatch-request r :log))))))


(defun parse-body (body-string &optional (delimiters '(#\&)))
  "Parse BODY-STRING returning list of (var val) pairs"
  (mapcar (lambda (x)
            (mapcar #'urlstring-unescape (split x 2 "=")))
          (split body-string nil delimiters)))

(defun read-folded-line (stream &optional eof-error-p eof-value)
  "_Should_ read a complete logical header line, including folding
continuation lines.  Doesn't work yet."
  (let ((a (string-right-trim `(#\Newline #\Linefeed ,(code-char 13))
			      (read-line stream eof-error-p eof-value))))
    ;;(format t "Got: ~A~%" a)
    a))


;;; Unless you're me (and probably not even then), you should be
;;; using REQUEST-HEADER not this:

(defun header-value (name header-list)
  "Get the value of the header named NAME in HEADER-LIST"
  (cadr (assoc name header-list)))

(defun read-headers (stream)
  (let ((headers nil))
    (do ((line (read-folded-line stream t) (read-folded-line stream t)))
        ((or (not line) (string= line "")) headers)
      (destructuring-bind (keyword value) (split line 2 '(#\:))
        ;; RFC 1945: "Each header field consists of a name followed
        ;; immediately by a colon (":"), a single space (SP)
        ;; character, and the field value. Field names are case-
        ;; insensitive." - so we need to lose the first char of value
        (let* ((value (subseq value 1 nil))
               (keyword-symbol (intern (string-upcase keyword) :keyword))
               (keyword-value (assoc keyword-symbol  headers)))
          (if keyword-value
              (setf (cdr keyword-value)
                    (append (cdr keyword-value) (list value)))
            (setf headers (acons keyword-symbol (list value) headers))))))
    headers))
        
